/*
 * autosort.c: A channel autosort plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 */

#include <vdr/plugin.h>
#include <ctype.h>
#include <vdr/sources.h>
#include <vdr/channels.h>
#include <vdr/interface.h>
#include <math.h>

// ToDo in 0.0.8
// Menu: Sort by default
// Option: NoSort
// Option: OpenCa/OpenProvider

static const char *VERSION        = "0.0.8";
static const char *DESCRIPTION    = "Channel Autosort";
static const char *MAINMENUENTRY  = "AutoSort Config Test";

// -----------------------------------------------------
int maxMoves = 20;
int messageLevel = 2;
// 0 - No Messages
// 1 - LoopsReady
// 2 - GroupReady
// 3 - ChannelMove
int messageTime = 5;
int lastTimeStampSet = 0;
int timeStampSetDelta = 604800;
// 1 hour   = 3600
// 1 day    = 86400
// 1 week   = 604800
// 1 month  = 2592000
bool writeNewTimeStamp = false;
bool writeTimeStamp = true;
bool sortUnsorted = true;
bool sortNew = false;
bool seperateNew = false;
bool hideMenu = true;
bool housekeepingRun = true;
bool versionSorting = true;
bool saveSetup = false;
bool useFixedAuto = false;
bool useFixedUnsorted = false;
bool useFixedNew = false;
int fixedAuto = 0;
int fixedUnsorted = 0;
int fixedNew = 0;
// -----------------------------------------------------

char *UpString (char *ToUp) {
  int i = 0;
  char *Up = ToUp;  
  while (ToUp[i] != '\0'){
    Up[i] = toupper(ToUp[i]);
    i++;
    }  
  return Up;
}

char *KillSpecial(char *s)
{
  int i = 0;
  int j = 0;
  char *ks = s;  
  while (s[i] != '\0'){
    if ((!isspace(s[i])) && (!ispunct(s[i]))) {
      ks[j] = s[i];
      j++;
      }
    i++;
    } 
    ks[j] = '\0';
  return ks;
}

// -----------------------------------------------------

class cAutoGroup : public cListObject {
private:
  char *name;
  int priority;
  int pIdx;
  int order;
  int oIdx;
  int fixedPos;  
  char *startName;
  struct t_hasName{
    char *hasName;
    };
  t_hasName hasNameList[10];
  struct t_notName{
    char *notName;
    };
  t_notName notNameList[10];
  struct t_hasSource{
    char *hasSource;
    };
  t_hasSource hasSourceList[10];
  struct t_hasProvider{
    char *hasProvider;
    };
  t_hasProvider hasProviderList[10];    
  struct t_notProvider{
    char *notProvider;
    };
  t_notProvider notProviderList[10];    
  struct t_caValue{
    char *caValue;
    };
  t_caValue caValueList[10];    
  struct t_notCaValue{
    char *notCaValue;
    };
  t_notCaValue notCaValueList[10];    
  struct t_audioLang{
    char *audioLang;
    };
  t_audioLang audioLangList[10];    
  bool subSections;
  bool sort;
  bool keepHere;
  bool setCaFF;
  bool setCa0;
  bool isSpecial;
  // Tri-states
  // -1 : must NOT have
  //  0 : don't care
  //  1 : must have
  int hasApid;
  int hasVpid;
  int hasTpid;
  int hasCa;
  int hasOpenCa;
  int hasClosedCa;
  int timeStampDays;
public:
  cAutoGroup(void);
  virtual ~cAutoGroup();
  cAutoGroup *Copy(cAutoGroup *OldGroup);
  bool Parse(const char *s);
  void Print(void);
  char *Name(void)                       { return name; };
  void SetName(char *Name)               { name = strdup(Name); };
  char *StartName(void)                  { return startName; };
  void SetStartName(char *StartName)     { startName = strdup(StartName); };
  int Priority (void)                    { return priority; };  
  void SetPriority (int Priority)        { priority = Priority; }; 
  int Order (void)                       { return order; };  
  int OIdx (void)                        { return oIdx; };  
  int PIdx (void)                        { return pIdx; };  
  bool KeepHere (void)                   { return keepHere; };
  bool Sort (void)                       { return sort; };
  bool SubSections (void)                { return subSections; };
  void SetSubSections (bool SubSections) { subSections = SubSections; };
  bool IsSpecial (void)                  { return isSpecial; };
  bool SetCaFF (void)                    { return setCaFF; };
  bool SetCa0 (void)                     { return setCa0; };
  void SetOrder (int Order)              { order = Order; };
  void SetFixedPos (int FPos)            { fixedPos = FPos; };
  int FixedPos (void)                    { return fixedPos; };
  void SetPIdx (int PIdx)                { pIdx = PIdx; }; 
  void SetOIdx (int OIdx)                { oIdx = OIdx; }; 
  void SetSort (void)                    { sort = true; }; 
  void SetSpecial (void)                 { isSpecial = true; };
  int HasApid (void)                     { return hasApid; };
  int HasVpid (void)                     { return hasVpid; };
  int HasTpid (void)                     { return hasTpid; };
  int HasCa (void)                       { return hasCa; };
  int HasOpenCa (void)                   { return hasOpenCa; };
  int HasClosedCa (void)                 { return hasClosedCa; };
  int TimeStampDays (void)               { return timeStampDays; };
  char *HasName (int NIdx)               { return hasNameList[NIdx].hasName; };
  char *NotName (int nNIdx)              { return notNameList[nNIdx].notName; };
  char *HasSource (int SIdx)             { return hasSourceList[SIdx].hasSource; };
  char *HasProvider (int PrIdx)          { return hasProviderList[PrIdx].hasProvider; };
  char *NotProvider (int PrIdx)          { return notProviderList[PrIdx].notProvider; };
  char *CaValue (int CaIdx)              { return caValueList[CaIdx].caValue; };
  char *NotCaValue (int nCaIdx)          { return notCaValueList[nCaIdx].notCaValue; };
  char *AudioLang (int AIdx)             { return audioLangList[AIdx].audioLang; };
  int GetChIndex(void);
  };

cAutoGroup::cAutoGroup(void)
{
  pIdx = 0;
  oIdx = 0;
  name = strdup("");
  startName = strdup("");
  hasNameList[0].hasName         = strdup("NotSet");
  notNameList[0].notName         = strdup("NotSet");
  hasSourceList[0].hasSource     = strdup("NotSet");
  hasProviderList[0].hasProvider = strdup("NotSet");
  notProviderList[0].notProvider = strdup("NotSet");
  caValueList[0].caValue         = strdup("NotSet");
  notCaValueList[0].notCaValue   = strdup("NotSet");
  audioLangList[0].audioLang     = strdup("NotSet");
  priority = 99;
  order = 500;
  fixedPos = -1;
  subSections = false;
  sort = false;
  keepHere = false;
  setCaFF = false;
  setCa0 = false;
  isSpecial = false;
  hasApid = 0;
  hasVpid = 0;
  hasTpid = 0;
  hasCa = 0;
  hasOpenCa = 0;
  hasClosedCa = 0;
  timeStampDays = 0;
}

cAutoGroup *cAutoGroup::Copy(cAutoGroup *OldGroup)
{
  cAutoGroup *NewGroup = new cAutoGroup;
  int i;
  name        = OldGroup->Name();
  priority    = OldGroup->Priority();
  order       = OldGroup->Order();
  fixedPos    = OldGroup->FixedPos();
  subSections = OldGroup->SubSections();
  sort        = OldGroup->Sort();
  keepHere    = OldGroup->KeepHere();
  setCaFF     = OldGroup->SetCaFF();
  setCa0      = OldGroup->SetCa0();
  isSpecial   = OldGroup->IsSpecial();
  hasApid     = OldGroup->HasApid();
  hasVpid     = OldGroup->HasVpid();
  hasTpid     = OldGroup->HasTpid();
  hasCa       = OldGroup->HasCa();
  hasOpenCa   = OldGroup->HasOpenCa();
  hasClosedCa = OldGroup->HasClosedCa();
  timeStampDays = OldGroup->TimeStampDays();
  i = 0; while (strcmp(OldGroup->HasName(i),     "NotSet") != 0) { hasNameList[i].hasName         = OldGroup->HasName(i);     i++; } hasNameList[i].hasName         = strdup("NotSet");
  i = 0; while (strcmp(OldGroup->NotName(i),     "NotSet") != 0) { notNameList[i].notName         = OldGroup->NotName(i);     i++; } notNameList[i].notName         = strdup("NotSet");
  i = 0; while (strcmp(OldGroup->HasSource(i),   "NotSet") != 0) { hasSourceList[i].hasSource     = OldGroup->HasSource(i);   i++; } hasSourceList[i].hasSource     = strdup("NotSet");
  i = 0; while (strcmp(OldGroup->HasProvider(i), "NotSet") != 0) { hasProviderList[i].hasProvider = OldGroup->HasProvider(i); i++; } hasProviderList[i].hasProvider = strdup("NotSet");
  i = 0; while (strcmp(OldGroup->NotProvider(i), "NotSet") != 0) { notProviderList[i].notProvider = OldGroup->NotProvider(i); i++; } notProviderList[i].notProvider = strdup("NotSet");
  i = 0; while (strcmp(OldGroup->CaValue(i),     "NotSet") != 0) { caValueList[i].caValue         = OldGroup->CaValue(i);     i++; } caValueList[i].caValue         = strdup("NotSet");
  i = 0; while (strcmp(OldGroup->NotCaValue(i),  "NotSet") != 0) { notCaValueList[i].notCaValue   = OldGroup->NotCaValue(i);  i++; } notCaValueList[i].notCaValue   = strdup("NotSet");
  i = 0; while (strcmp(OldGroup->AudioLang(i),   "NotSet") != 0) { audioLangList[i].audioLang     = OldGroup->AudioLang(i);   i++; } audioLangList[i].audioLang     = strdup("NotSet");
  return NewGroup;
}

cAutoGroup::~cAutoGroup()
{
}

int cAutoGroup::GetChIndex(void)
{
  cChannel *channel = Channels.Last();
  while (channel){
    if ((strcmp(channel->Name(),name) == 0) && channel->GroupSep()) return channel->Index();
    if (strcmp(channel->Name(),"Auto Sort") == 0) return 0;
    channel = Channels.Get(Channels.GetPrevGroup(channel->Index()));
    }  
  return 0;
}

void cAutoGroup::Print(void)
{
  int i = 0;
  dsyslog("AutoGroupPrint:");
  dsyslog("  Name:          %s",name);
  dsyslog("  Index:         %d",Index());
  dsyslog("  Priority:      %d",priority);
  dsyslog("  Priority Index:%d",pIdx);
  dsyslog("  Order:         %d",order);
  dsyslog("  Order Index:   %d",oIdx);
  if (fixedPos < 0) { dsyslog("  FixedPosition: None");}
    else {            dsyslog("  FixedPosition: %d",fixedPos); }
  if (timeStampDays = 0) { dsyslog("  TimeStamp:     Dont't care about");}
    else {            dsyslog("  TimeStamp:     %d days old",timeStampDays); }
  if (subSections) {  dsyslog("  SubSections:   Yes"); }
    else {            dsyslog("  SubSections:   No"); }
  if (sort) {         dsyslog("  Sort:          Yes"); }
    else {            dsyslog("  Sort:          No"); }
  if (keepHere) {     dsyslog("  KeepHere:      Yes"); }
    else {            dsyslog("  KeepHere:      No"); }
  if (setCaFF) {      dsyslog("  SetCaFF:       Yes"); }
    else {            dsyslog("  SetCaFF:       No"); }
  if (setCa0) {       dsyslog("  SetCa0:        Yes"); }
    else {            dsyslog("  SetCa0:        No"); }
  if (hasApid  < 0) {     dsyslog("  Apid:          Must NOT have"); }
  if (hasApid == 0) {     dsyslog("  Apid:          Don't care about"); }
  if (hasApid  > 0) {     dsyslog("  Apid:          Must have"); }
  if (hasVpid  < 0) {     dsyslog("  Vpid:          Must NOT have"); }
  if (hasVpid == 0) {     dsyslog("  Vpid:          Don't care about"); }
  if (hasVpid  > 0) {     dsyslog("  Vpid:          Must have"); }
  if (hasTpid  < 0) {     dsyslog("  Tpid:          Must NOT have"); }
  if (hasTpid == 0) {     dsyslog("  Tpid:          Don't care about"); }
  if (hasTpid  > 0) {     dsyslog("  Tpid:          Must have"); }
  if (hasCa  < 0) {       dsyslog("  Ca:            Must NOT have"); }
  if (hasCa == 0) {       dsyslog("  Ca:            Don't care about"); }
  if (hasCa  > 0) {       dsyslog("  Ca:            Must have"); }
  if (hasOpenCa  < 0) {   dsyslog("  OpenCa:        Must NOT have"); }
  if (hasOpenCa == 0) {   dsyslog("  OpenCa:        Don't care about"); }
  if (hasOpenCa  > 0) {   dsyslog("  OpenCa:        Must have"); }  
  if (hasClosedCa  < 0) { dsyslog("  ClosedCa:      Must NOT have"); }
  if (hasClosedCa == 0) { dsyslog("  ClosedCa:      Don't care about"); }
  if (hasClosedCa  > 0) { dsyslog("  ClosedCa:      Must have"); }
  if (!strcmp(startName,"")) { dsyslog("  StartName:     Don't care about");}
    else {                     dsyslog("  StartName:     %s",startName); }
  i = 0; while (strcmp(hasNameList[i].hasName,"NotSet") != 0) {
                          dsyslog("  HasName[%d]     %s",i ,hasNameList[i].hasName); i++; }
  i = 0; while (strcmp(notNameList[i].notName,"NotSet") != 0) {
                          dsyslog("  NotName[%d]     %s",i ,notNameList[i].notName); i++; }
  i = 0; while (strcmp(hasSourceList[i].hasSource,"NotSet") != 0) {
                          dsyslog("  HasSource[%d]   %s",i ,hasSourceList[i].hasSource); i++; }
  i = 0; while (strcmp(hasProviderList[i].hasProvider,"NotSet") != 0) {
                          dsyslog("  HasProvider[%d] %s",i ,hasProviderList[i].hasProvider); i++; }
  i = 0; while (strcmp(notProviderList[i].notProvider,"NotSet") != 0) {
                          dsyslog("  NotProvider[%d] %s",i ,notProviderList[i].notProvider); i++; }
  i = 0; while (strcmp(caValueList[i].caValue,"NotSet") != 0) {
                          dsyslog("  CaValue[%d]     %s",i ,caValueList[i].caValue); i++; }
  i = 0; while (strcmp(notCaValueList[i].notCaValue,"NotSet") != 0) {
                          dsyslog("  NotCaValue[%d]  %s",i ,notCaValueList[i].notCaValue); i++; }
  i = 0; while (strcmp(audioLangList[i].audioLang,"NotSet") != 0) {
                          dsyslog("  AudioLang[%d]   %s",i ,audioLangList[i].audioLang); i++; }
}

bool cAutoGroup::Parse(const char *line)
{
  char *subline;
  subline = strdup(line);
  char *option = strtok(subline,":");
  bool nameFound = false;
  int i = 0;
  while (option != NULL)
  {
    if      (!nameFound) { name = option; nameFound = true; }
    else if (!strcasecmp(option, "HasApid"))          hasApid =  1;
    else if (!strcasecmp(option, "HasNoApid"))        hasApid = -1; 
    else if (!strcasecmp(option, "HasVpid"))          hasVpid =  1; 
    else if (!strcasecmp(option, "HasNoVpid"))        hasVpid = -1; 
    else if (!strcasecmp(option, "HasTpid"))          hasTpid =  1; 
    else if (!strcasecmp(option, "HasNoTpid"))        hasTpid = -1; 
    else if (!strcasecmp(option, "HasCa"))            hasCa =  1; 
    else if (!strcasecmp(option, "HasNoCa"))          hasCa = -1; 
    else if (!strcasecmp(option, "HasOpenCa"))        hasOpenCa =  1; 
    else if (!strcasecmp(option, "HasNoOpenCa"))      hasOpenCa = -1; 
    else if (!strcasecmp(option, "HasClosedCa"))      hasClosedCa =  1; 
    else if (!strcasecmp(option, "HasNoClosedCa"))    hasClosedCa = -1; 
    else if (!strcasecmp(option, "SubSections"))      subSections = true; 
    else if (!strcasecmp(option, "Sort"))             sort = true; 
    else if (!strcasecmp(option, "keepHere"))         keepHere = true; 
    else if (!strcasecmp(option, "SetCaFF"))          setCaFF = true; 
    else if (!strcasecmp(option, "SetCa0"))           setCa0 = true; 
    else if (!strncasecmp(option, "Priority",8))    { strsep(&option,"="); priority = atoi(option); } 
    else if (!strncasecmp(option, "Order",5))       { strsep(&option,"="); order =    atoi(option); } 
    else if (!strncasecmp(option, "FixedPos",8))    { strsep(&option,"="); fixedPos = atoi(option); } 
    else if (!strncasecmp(option, "TimeStamp",9))   { strsep(&option,"="); timeStampDays = atoi(option); } 
    else if (!strncasecmp(option, "StartName",9))   { strsep(&option,"="); startName = option; } 
    else if (!strncasecmp(option, "HasName",7))     { strsep(&option,"="); i = 0;
                                                      while (strcmp(hasNameList[i].hasName,"NotSet")) { i++; }
                                                      hasNameList[i  ].hasName = strdup(UpString(option));
                                                      hasNameList[i+1].hasName = strdup("NotSet"); }
    else if (!strncasecmp(option, "NotName",7))     { strsep(&option,"="); i = 0;
                                                      while (strcmp(notNameList[i].notName,"NotSet")) { i++; }
                                                      notNameList[i  ].notName = strdup(UpString(option));
                                                      notNameList[i+1].notName = strdup("NotSet"); }
    else if (!strncasecmp(option, "HasSource",9))   { strsep(&option,"="); i = 0;
                                                      while (strcmp(hasSourceList[i].hasSource,"NotSet")) { i++; }
                                                      hasSourceList[i  ].hasSource = strdup(UpString(option));
                                                      hasSourceList[i+1].hasSource = strdup("NotSet"); }
    else if (!strncasecmp(option, "HasProvider",11)) { strsep(&option,"="); i = 0;
                                                      while (strcmp(hasProviderList[i].hasProvider,"NotSet")) { i++; }
                                                      hasProviderList[i  ].hasProvider = strdup(UpString(option));
                                                      hasProviderList[i+1].hasProvider = strdup("NotSet"); }
    else if (!strncasecmp(option, "NotProvider",11)) { strsep(&option,"="); i = 0;
                                                      while (strcmp(notProviderList[i].notProvider,"NotSet")) { i++; }
                                                      notProviderList[i  ].notProvider = strdup(UpString(option));
                                                      notProviderList[i+1].notProvider = strdup("NotSet"); }
    else if (!strncasecmp(option, "CaValue",7))     { strsep(&option,"="); i = 0;
                                                      while (strcmp(caValueList[i].caValue,"NotSet")) { i++; }
                                                      caValueList[i  ].caValue = strdup(UpString(option));
                                                      caValueList[i+1].caValue = strdup("NotSet"); }
    else if (!strncasecmp(option, "NotCaValue",10)) { strsep(&option,"="); i = 0;
                                                      while (strcmp(notCaValueList[i].notCaValue,"NotSet")) { i++; }
                                                      notCaValueList[i  ].notCaValue = strdup(UpString(option));
                                                      notCaValueList[i+1].notCaValue = strdup("NotSet"); }
    else if (!strncasecmp(option, "AudioLang",9))   { strsep(&option,"="); i = 0;
                                                      while (strcmp(audioLangList[i].audioLang,"NotSet")) { i++; }
                                                      audioLangList[i  ].audioLang = strdup(UpString(option));
                                                      audioLangList[i+1].audioLang = strdup("NotSet"); }
    option = strtok(NULL,":");
  }
  //Print();
  if (!strcasecmp(name,"Auto Sort") ||
      !strcasecmp(name,"Unsorted") ||
      !strcasecmp(name,"Dummy") ||
      !strcasecmp(name,"Dummy2") ||
      !strcasecmp(name,"New Channels")) {
	esyslog("AutoSort -ERROR-  Found unallowed AutoGroup name: '%s'",name);
	esyslog("       !!! Check your autosort.conf !!!");
        return false;
	}
  else if (priority < 0 || priority > 99  ||
           order    < 1 || order    > 999 ||
           fixedPos < -1 ) {
	     esyslog("AutoSort -ERROR-  Found wrong Priority, Order or FixedPos in AutoGroup '%s'",name);
	     esyslog("       !!! Check your autosort.conf !!!");
             return false;     
             }
  else if (setCaFF && setCa0) {
	esyslog("AutoSort -ERROR-  Found SetCaFF and SetCa0 in same AutoGroup '%s'",name);
	esyslog("       !!! Check your autosort.conf !!!");
        return false;
	}  
  else return true;
}

// -----------------------------------------------------

class cAutoGroups : public cConfig<cAutoGroup> {
private:
  void ReIndex(void);
public:
  cAutoGroups(void);
  bool Load (const char *FileName, bool AllowComments, bool MustExist);
  cAutoGroup *GetFirstByOrder (void);
  cAutoGroup *GetNextByOrder (int OrdIdx);
  cAutoGroup *GetFirstByPrio (void);
  cAutoGroup *GetNextByPrio (int PrioIdx);
  cAutoGroup *GetByName (char *search);
  };

cAutoGroups AutoGroups;

cAutoGroups::cAutoGroups(void)
{
}

bool cAutoGroups::Load(const char *FileName, bool AllowComments, bool MustExist)
{
  if (cConfig<cAutoGroup>::Load(FileName, AllowComments, MustExist)) {
    // Check for double Group Names
    for (cAutoGroup *group = First(); group; group = Next(group)) {
      for (cAutoGroup *group2 = Next(group); group2; group2 = Next(group2)) {
        if (!strcasecmp(group->Name(),group2->Name())){
	  esyslog("AutoSort -ERROR-  AutoGroup '%s' is double",group->Name());
	  group2->SetName(strcat(group2->Name()," -Double-"));
	  esyslog("AutoSort -ERROR-  second AutoGroup renamed to '%s'",group2->Name());
	  esyslog("       !!! Check your autosort.conf !!!");
	  }
        }
      }
    // Expand Subsection groups
    char *subChar;    
    for (cAutoGroup *group = First(); group; group = Next(group)) {
      if (group->SubSections()) {
        dsyslog("AutoSort: Expanding SubSection group %s", group->Name());
	for (int i = 90; i >= 65; i--) {
          subChar="";	
	  asprintf(&subChar, "%c", i);	
	  cAutoGroup *SubGroup = new cAutoGroup;
	  SubGroup->Copy(group);
	  SubGroup->SetName(strcat(strdup(group->Name())," "));
	  SubGroup->SetName(strcat(SubGroup->Name(),strdup(subChar)));
	  SubGroup->SetPriority(group->Priority() - 1);
	  SubGroup->SetOrder(group->Order() + 1);
	  SubGroup->SetSubSections(false);
	  SubGroup->SetFixedPos( -1 );
	  SubGroup->SetStartName(subChar);
          Add(SubGroup, group);
	  }
        }
      }
    // Add some special groups
    Add(new cAutoGroup);
    cAutoGroup *special = Last();
    special->SetSpecial(); special->SetName("Auto Sort"); special->SetOrder(0);
    if (useFixedAuto) special->SetFixedPos(fixedAuto);
    Add(new cAutoGroup);
    special = Last();
    special->SetSpecial(); special->SetName("Unsorted"); special->SetOrder(1010);
    if (sortUnsorted) special->SetSort(); 
    if (useFixedUnsorted) special->SetFixedPos(fixedUnsorted);
    if (seperateNew){ 
      Add(new cAutoGroup);
      special = Last();
      special->SetSpecial(); special->SetName("New Channels"); special->SetOrder(1020); 
      if (sortNew) special->SetSort();
      if (useFixedNew) special->SetFixedPos(fixedNew);
      }
    ReIndex();
    return true;
    }
  esyslog("Auto Sort: -ERROR- Can't find %s/autosort.conf -EXITING-",cPlugin::ConfigDirectory());
  esyslog("       !!! Check your autosort.conf !!!");
  return false;
}

void cAutoGroups::ReIndex( void )
{
  struct t_map {
    int prio;
    int ord;
    int idx;
    };
  t_map map[Count()];
  t_map dummy;
  bool ready = false;
  for (cAutoGroup *group = First(); group; group = Next(group)) {
    map[group->Index()].idx = group->Index();
    map[group->Index()].prio = group->Priority();
    map[group->Index()].ord = group->Order();
    }
  // Sort by Priority
  while (!ready) { 
    ready = true;    
    for (int n = 0; n < Count(); n++) {
      if ( n + 1 < Count()) {
        if (map[n].prio > map[n+1].prio){
	  ready = false;
	  dummy = map[n+1];
	  map[n+1] = map[n];
	  map[n] = dummy;
	  }
        }
      }
    } // while
    for (int n = 0; n < Count(); n++) {
      cAutoGroup *group = Get(map[n].idx);
      group->SetPIdx(n);
      }            
  ready = false;
  // Sort by Order
  while (!ready) { 
    ready = true;    
    for (int n = 0; n < Count(); n++) {
      if ( n + 1 < Count()) {
        if (map[n].ord > map[n+1].ord){
	  ready = false;
	  dummy = map[n+1];
	  map[n+1] = map[n];
	  map[n] = dummy;
	  }
        }
      }
    } // while
    for (int n = 0; n < Count(); n++) {
      cAutoGroup *group = Get(map[n].idx);
      group->SetOIdx(n);
      }
}

cAutoGroup *cAutoGroups::GetByName (char *search){
  for (cAutoGroup *group = First(); group; group = Next(group)) {
    if (!strcmp(search,group->Name())) return group;
    }
  return NULL;
}

cAutoGroup *cAutoGroups::GetFirstByOrder (void){
  for (cAutoGroup *group = First(); group; group = Next(group)) {
    if (group->OIdx() == 0) return group;
    }
  return NULL;
}

cAutoGroup *cAutoGroups::GetNextByOrder (int OrdIdx){
  OrdIdx++;
  for (cAutoGroup *group = First(); group; group = Next(group)) {
    if (group->OIdx() == OrdIdx ) return group;
    }
  return NULL;
}

cAutoGroup *cAutoGroups::GetFirstByPrio (void){
  for (cAutoGroup *group = First(); group; group = Next(group)) {
    if (group->PIdx() == 0) return group;
    }
  return NULL;
}

cAutoGroup *cAutoGroups::GetNextByPrio (int PrioIdx){
  PrioIdx++;
  for (cAutoGroup *group = First(); group; group = Next(group)) {
    if (group->PIdx() == PrioIdx ) return group;
    }
  return NULL;
}

// ------------------------------------------

class cNewAutoSort {
private:
  cAutoGroup *workingGroup;
public:
  cNewAutoSort(void);
  virtual ~cNewAutoSort();
  void Dump(int from, int to);
  bool UserBreak (bool CompleteRun = false);
  void CheckDelimiters (void);
  void CheckChannels (bool completeRun = false);
  void Sort(void);
  bool ChannelFitsGroup (int ChIndex, cAutoGroup *group);
  void MoveChannelToGroup (int ChIndex, cAutoGroup *group);
  void MoveGroupToEnd (int ChIndex);
  void MoveGroupBehindGroup (int ChIndex, cAutoGroup *group);
  void AddGroup (cAutoGroup *group);
  void InsGroup (cAutoGroup *group, int ChIndex);
  void DeleteGroup (int ChIndex);
  void ModifyChannelToGroup (int ChIndex, cAutoGroup *group, bool Out = false);
  void SetTimeStamp (int ChIndex);
  const char *InGroup (int ChIndex);
  };

cNewAutoSort::cNewAutoSort(void)
{
  AutoGroups.Load(AddDirectory(cPlugin::ConfigDirectory(), "autosort.conf"), true, true);
  CheckDelimiters();
  workingGroup = AutoGroups.GetFirstByPrio();
  if ( (time(NULL) - timeStampSetDelta) > lastTimeStampSet) writeNewTimeStamp = true;
  if (0){ // ToDo: replace by config option "On Startup"
    CheckChannels();
    }
}

cNewAutoSort::~cNewAutoSort()
{
}

void cNewAutoSort::Sort(void)
{
  cAutoGroup *group = workingGroup;
  int firstChIndex;
  int lastChIndex;
  int moves = 0;
  //while (group) {
    if (group->Sort()){
      firstChIndex = group->GetChIndex() + 1;
      //if (AutoGroups.GetNextByOrder(group->OIdx())){
        //lastChIndex = AutoGroups.GetNextByOrder(group->OIdx())->GetChIndex() - 1;
      if (Channels.GetNextGroup(group->GetChIndex())){ // XXX
        lastChIndex = Channels.GetNextGroup(group->GetChIndex()) - 1;
      } else {
	lastChIndex = Channels.Last()->Index();
	}
      if ((lastChIndex - firstChIndex) > 0){
        //dsyslog("AutoSort: sorting group %s",group->Name());
	for (cChannel *channel1 = Channels.Get(firstChIndex); channel1->Index() < lastChIndex; channel1 = Channels.Next(channel1)) {
	  int ch2Index = channel1->Index() +1;
	  if (versionSorting) {
	    while ( (ch2Index <= lastChIndex) && (strverscmp(UpString(KillSpecial(strdup(channel1->Name()))), UpString(KillSpecial(strdup(Channels.Get(ch2Index)->Name())))) > 0 ) ) {
	      ch2Index++;
	      }
	  } else {
	    while ( (ch2Index <= lastChIndex) && (strcasecmp(channel1->Name(), Channels.Get(ch2Index)->Name()) > 0 ) ) {
	      ch2Index++;
	      }
	    }
	  ch2Index--;
	  if (channel1->Index() != ch2Index ){
	    if (channel1->Index() < AutoGroups.GetFirstByOrder()->GetChIndex() ){
	      dsyslog("AutoSort: -ERROR- ::Sort() trys working before Auto-Sort-Mark  -ABORTED-");
	      return;
	      }	    
	    dsyslog("AutoSort: sort channel '%s' behind '%s'", channel1->Name(), Channels.Get(ch2Index)->Name());
	    if (ch2Index != Channels.Last()->Index()) { 
	      Channels.Move(channel1, Channels.Get(ch2Index));
              Channels.ReNumber();
              Channels.SetModified();
	      if (strcmp(InGroup(ch2Index), group->Name())!=0) {
                dsyslog ("ERROR: (Sort) Moved %s to %s instead of %s ", Channels.Get(ch2Index)->Name(), InGroup(ch2Index), group->Name());
                Dump(ch2Index -3,ch2Index +3);
                }
	      moves++;
	      if (moves > maxMoves) return;
	      }
            }
	  }
	}
      }
    //group = AutoGroups.GetNextByOrder(group->OIdx());
    //}
}

//#undef DELETECA

void cNewAutoSort::ModifyChannelToGroup (int ChIndex, cAutoGroup *group, bool Out)
{
  cChannel *channel = Channels.Get(ChIndex);
  const int CaFF[] = {255,0};
#ifndef DELETECA
  const int Ca0[] = {0};
#endif
  if (!Out) { // channel will go into group
    if (group->SetCaFF()) {
#ifdef DELETECA
      channel->DeleteCaIds();
#else      
      if ((channel->Ca(0) >= 0x0001) && (channel->Ca(0) <= 0x00FF)){
        isyslog("AutoSort: -WARNING- You need the 'deleteca'-patch for '%s'", channel->Name());
      } else {
        channel->SetCaIds(Ca0);      
        }
#endif      
      channel->SetCaIds(CaFF);
      }
    if (group->SetCa0()) {
#ifdef DELETECA
      channel->DeleteCaIds();
#else      
      if ((channel->Ca(0) >= 0x0001) && (channel->Ca(0) <= 0x00FF)){
        isyslog("AutoSort: -WARNING- You need the 'deleteca'-patch for '%s'", channel->Name());
      } else {
        channel->SetCaIds(Ca0);      
        }
#endif      
      }
  } else { // channel will go out of group
    if (group->SetCaFF()) {
#ifdef DELETECA
      channel->DeleteCaIds();
#else      
      if ((channel->Ca(0) >= 0x0001) && (channel->Ca(0) <= 0x00FF)){
        isyslog("AutoSort: -WARNING- You need the 'deleteca'-patch for '%s'", channel->Name());
      } else {
        channel->SetCaIds(Ca0);      
        }
#endif      
      }
    }
}

bool cNewAutoSort::ChannelFitsGroup (int ChIndex, cAutoGroup *group)
{
  cChannel *channel = Channels.Get(ChIndex);
  int i = 0;
  int j = 0;
  bool fits = false;
  // Tri-states
  // -1 : must NOT have
  //  0 : don't care
  //  1 : must have  
  if (( group->HasApid() > 0) && (channel->Apid(0) == 0 )) return false;
  if (( group->HasApid() < 0) && (channel->Apid(0) != 0 )) return false;  
  if (( group->HasVpid() > 0) && (channel->Vpid()  == 0 )) return false;
  if (( group->HasVpid() < 0) && (channel->Vpid()  != 0 )) return false;  
  if (( group->HasTpid() > 0) && (channel->Tpid()  == 0 )) return false;
  if (( group->HasTpid() < 0) && (channel->Tpid()  != 0 )) return false;
  if (( group->HasCa()   > 0) && (channel->Ca(0)   == 0 )) return false;
  if (( group->HasCa()   < 0) && (channel->Ca(0)   != 0 )) return false;

  if (strcmp(group->StartName(), "") != 0) {
    if (strncasecmp(group->StartName(), channel->Name(), 1) != 0) return false;
    }

  if (group->TimeStampDays() > 0) {
    if (strstr(channel->ShortName(),"ASTS")){
      char *buf = strdup(channel->ShortName());
      strsep(&buf,"-");
      int stamp = atoi(buf);
      if ((time(NULL) - stamp) < ( group->TimeStampDays() * 86400 )) return false;
    } else {
      return false; // no TimeStamp
      }
    }
    
  if (strcmp(group->HasName(0),"NotSet")){
    i = 0;
    fits = false;
    while (strcmp(group->HasName(i),"NotSet") != 0) {
      if (strstr(UpString(strdup(channel->Name())),group->HasName(i))) {
        fits = true;
        }
      i++;
      }
    if (!fits) return false;
    }

  if (strcmp(group->NotName(0),"NotSet")){
    i = 0;
    while (strcmp(group->NotName(i),"NotSet") != 0) {
      if (strstr(UpString(strdup(channel->Name())),group->NotName(i))) {
        fits = true; //ToDo: delete this ?
	return false;
        }
      i++;
      }
    }
  if (strcmp(group->HasSource(0),"NotSet")){
    i = 0;
    fits = false;
    while (strcmp(group->HasSource(i),"NotSet") != 0) {
      if (channel->Source() == cSource::FromString(group->HasSource(i))) {
        fits = true;
        }	
      i++;
      }
    if (!fits) return false;
    }

  if (strcmp(group->HasProvider(0),"NotSet")){
    i = 0;
    fits = false;
    while (strcmp(group->HasProvider(i),"NotSet") != 0) {
      if (strstr(UpString(strdup(channel->Provider())),group->HasProvider(i))) {
        fits = true;
        }
      i++;
      }
    if (!fits) return false;
    }
        
  if (strcmp(group->NotProvider(0),"NotSet")){
    i = 0;
    while (strcmp(group->NotProvider(i),"NotSet") != 0) {
      if (strstr(UpString(strdup(channel->Provider())),group->NotProvider(i))) {
	return false;
        }
      i++;
      }
    }

  char *caHex;  
  if (strcmp(group->CaValue(0),"NotSet") != 0){
    i = 0;
    fits = false;
    while (strcmp(group->CaValue(i),"NotSet") != 0) {
      j = 0;      
      while (channel->Ca(j) != 0) {
        caHex="";	
	asprintf(&caHex, "%X", channel->Ca(j));	
        if (strstr(strdup(caHex),group->CaValue(i))) {
          fits = true;
          }	  
	j++;
	}	
      i++;
      }
    if (!fits) return false;
    }

  if (strcmp(group->NotCaValue(0),"NotSet") != 0){
    i = 0;
    while (strcmp(group->NotCaValue(i),"NotSet") != 0) {
      j = 0;      
      while (channel->Ca(j) != 0) {
        caHex="";	
	asprintf(&caHex, "%X", channel->Ca(j));	
        if (strstr(strdup(caHex),group->NotCaValue(i))) {
          return false;
          }	  
	j++;
	}	
      i++;
      }
    }

  if (strcmp(group->AudioLang(0),"NotSet") != 0){
    i = 0;
    fits = false;
    while (strcmp(group->AudioLang(i),"NotSet") != 0) {
      j = 0;
      while (strcmp(channel->Alang(j),"") != 0) {
        if (strstr(UpString(strdup(channel->Alang(j))),group->AudioLang(i))) {
          fits = true;
          }
	j++;
	}
      i++;
      }
    if (!fits) return false;
    }
  
  return true;
}


void cNewAutoSort::CheckChannels ( bool completeRun )
{
  char *buf;
  //cAutoGroup *workingGroup = AutoGroups.GetFirstByPrio();
  
  if (completeRun) workingGroup = AutoGroups.GetFirstByPrio();
  //if (completeRun) workingGroup = AutoGroups.GetByName("Sport");

  cAutoGroup *FromGroup = AutoGroups.GetNextByPrio(workingGroup->PIdx());
  cChannel *channel1;
  int i;
  int moves = 0;

  if ((completeRun) && (messageLevel >= 1)) Skins.Message(mtStatus,"AutoSort: Checking groups", messageTime);

  while ((workingGroup) && ((moves < maxMoves) || (completeRun))){
    if (UserBreak(completeRun)) {
      if (messageLevel >= 2) Skins.Message(mtStatus,NULL,0);
      return;
      }
    dsyslog("AutoSort: following rules for group %s", workingGroup->Name());
    if (messageLevel >= 2) {
      asprintf(&buf, "AutoSort: Checking group '%s'", workingGroup->Name());
      Skins.Message(mtStatus, buf, messageTime);    
      }
    if ( !workingGroup->KeepHere() && !workingGroup->IsSpecial()) {
      //dsyslog("Check channels in group %s for re-sort", workingGroup->Name());
      channel1 = Channels.Get(workingGroup->GetChIndex());
      if (Channels.GetNextGroup(channel1->Index()) > 0){	
        i = Channels.GetNextGroup(channel1->Index()) - 1;
      } else {
        i = Channels.Last()->Index();
        }
      //while ((i > channel1->Index()) && ((moves < maxMoves) || (completeRun))){
      while ((i > workingGroup->GetChIndex()) && ((moves < maxMoves) || (completeRun))){
        if (UserBreak(completeRun)) {
          if (messageLevel >= 2) Skins.Message(mtStatus,NULL,0);
          return;
          }	  
	if ((writeTimeStamp) && (writeNewTimeStamp)) SetTimeStamp(i); // found no better place for this
        alarm(60);
        if (!ChannelFitsGroup(i,workingGroup)){	
	  ModifyChannelToGroup(i, workingGroup, true);
          cAutoGroup *ToGroup = AutoGroups.GetByName("Unsorted");
          MoveChannelToGroup(i, ToGroup);
          moves++;
          }
        i--;
        }
      }
    if (!workingGroup->IsSpecial()) {
      FromGroup = AutoGroups.GetNextByPrio(workingGroup->PIdx());
      //while (FromGroup &&    //Check here if higher Prio Group isn't allowed to override KeepHere
      //      (strcmp(FromGroup->Name(),"New Channels") != 0) &&
      //      (strcmp(FromGroup->Name(),"Auto Sort") != 0)){ 
      while ((FromGroup) && (strcmp(FromGroup->Name(),"New Channels") != 0)){
        if (UserBreak(completeRun)) {
          if (messageLevel >= 2) Skins.Message(mtStatus,NULL,0);
          return;
          }
        //dsyslog("Check channels in group %s for group %s",FromGroup->Name(), workingGroup->Name());
        channel1 = Channels.Get(FromGroup->GetChIndex());
	if (Channels.GetNextGroup(channel1->Index()) > 0){	
	  i = Channels.GetNextGroup(channel1->Index()) - 1;
        } else {
	  i = Channels.Last()->Index();
          }
       while ((i > channel1->Index()) && ((moves < maxMoves) || (completeRun))){
         if (UserBreak(completeRun)) {
           if (messageLevel >= 2) Skins.Message(mtStatus,NULL,0);
           return;
           }
         alarm(60);
         if (ChannelFitsGroup(i,workingGroup)){
           if ((writeTimeStamp) && (writeNewTimeStamp)) SetTimeStamp(i); // found no better place for this
	   ModifyChannelToGroup(i, FromGroup, true);
	   ModifyChannelToGroup(i, workingGroup, false);
	   MoveChannelToGroup(i, workingGroup);
	   moves++;	   	   
	   }
	 i--;
         }      
        FromGroup = AutoGroups.GetNextByPrio(FromGroup->PIdx());
        }	
      }               
    if (messageLevel >= 2) Skins.Message(mtStatus,NULL,0);
    if ((moves >= maxMoves) && !completeRun) return;
    Sort();
    workingGroup = AutoGroups.GetNextByPrio(workingGroup->PIdx());
    if (messageLevel >= 1) Skins.Message(mtStatus,NULL,0);
    if (!workingGroup){ // We have all done
      if (messageLevel >= 1) Skins.Message(mtInfo,"AutoSort: Run Completed", messageTime);
      workingGroup = AutoGroups.GetFirstByPrio();
      if (writeNewTimeStamp) {
        //dsyslog("Run complete - timestamps wrote - i wish to save my setup");
        lastTimeStampSet = time(NULL);
	saveSetup = true;
	writeNewTimeStamp = false;
        }
      if ( time(NULL) - timeStampSetDelta > lastTimeStampSet) writeNewTimeStamp = true;
      return;
      }
    if (!completeRun) return; // Just one Group per Run
    if (UserBreak(completeRun)) return;
    }
}

bool cNewAutoSort::UserBreak (bool CompleteRun)
{
  eKeys key = cRemote::Get(0);
  if (NORMALKEY(key) == kNone) {
    return false;
  } else {
    if (!CompleteRun) {
      //dsyslog("UserBreak on Housekeeping");
      return true;
    } else {
      if (NORMALKEY(key) == kBack) {
        //dsyslog("UserBreak on CompleteRun");
        return true;
	}
      }
    }
  return false;
}

void cNewAutoSort::CheckDelimiters (void)
{
  if (messageLevel >= 1) Skins.Message(mtStatus,"AutoSort: Checking Group Delimiters", messageTime);
  cAutoGroup *group = AutoGroups.GetFirstByOrder();
  char *buf;  
  bool foundStart = false;
  int firstGroupChIdx = 1;
  for (cChannel *channel = Channels.Get(Channels.GetNextGroup(1)); (channel && !foundStart); channel = Channels.Get(Channels.GetNextGroup(channel->Index()))) 
    {
    if (strcmp(channel->Name(),group->Name()) == 0) 
      {
      foundStart = true;
      firstGroupChIdx = channel->Index();
      dsyslog("AutoSort: found group '%s' @ Index %d ",group->Name(), channel->Index());
      }
    }
  if (!foundStart){
    // Create all New
  if (messageLevel >= 1) Skins.Message(mtStatus,"AutoSort: Creating Group Delimiters", messageTime);
    group = AutoGroups.GetFirstByOrder();    
    while (group) {
      if (messageLevel >= 2) {
        asprintf(&buf, "AutoSort: Creating group '%s'", group->Name());
        Skins.Message(mtStatus, buf, messageTime);    
        }
      AddGroup(group);
      group= AutoGroups.GetNextByOrder(group->OIdx());
      }
  } else {
    // Is everybody there?
    if (messageLevel >= 1) Skins.Message(mtStatus,"AutoSort: Checking for missing Group Delimiters", messageTime);
    group = AutoGroups.GetFirstByOrder();    
    // Test for fixed Pos 'Auto Sort'
    if (group->FixedPos() > 0){
      if (group->FixedPos() != Channels.Get(firstGroupChIdx)->Number()){
        dsyslog ("AutoSort: Changing FixedPosition of 'Auto Sort' to %d", group->FixedPos());
        Channels.Get(firstGroupChIdx)->SetNumber(group->FixedPos());
        }
      }
    bool foundNext;
    while ((group = AutoGroups.GetNextByOrder(group->OIdx()))){
      foundNext = false;
      //dsyslog("AutoSort: Searching Group %s", group->Name());
      for (cChannel *channel = Channels.Get(Channels.GetNextGroup(firstGroupChIdx)); (channel && !foundNext); channel = Channels.Get(Channels.GetNextGroup(channel->Index()))) 
        {
        if (strcmp(channel->Name(),group->Name()) == 0) 
          {
          foundNext = true;
	  // Test for fixed Pos
	  if (group->FixedPos() > 0){
	    if (group->FixedPos() != channel->Number()){
	      dsyslog ("AutoSort: Changing FixedPosition of '%s' to %d",channel->Name(), group->FixedPos());
	      channel->SetNumber(group->FixedPos());
	      }
	    }	  
          }
	}  
        if (!foundNext) {
          if (messageLevel >= 2) {
            asprintf(&buf, "AutoSort: Creating group '%s'", group->Name());
            Skins.Message(mtStatus, buf, messageTime);    
            }
	  //dsyslog("AutoSort: Creating group '%s'", group->Name());
	  AddGroup(group);	  
	  }
      } 
    // Is anyone to much?    
    if (messageLevel >= 1) Skins.Message(mtStatus,"AutoSort: Checking for old Group Delimiters", messageTime);    
    for (cChannel *channel = Channels.Get(Channels.GetNextGroup(firstGroupChIdx)); channel; channel = Channels.Get(Channels.GetNextGroup(channel->Index()))) 
      {      
      foundNext = false;
      //dsyslog("AutoSort: Found Group %s", channel->Name());
      group = AutoGroups.GetFirstByOrder();    
      while ((group = AutoGroups.GetNextByOrder(group->OIdx()))){
        //dsyslog("AutoSort: Compare with %s",group->Name());
        if (strcmp(channel->Name(),group->Name()) == 0) 
          {
          foundNext = true;
          }	
        }
        if (!foundNext) {
	  int saveIndex = channel->Index();
          if (messageLevel >= 2) {
            asprintf(&buf, "AutoSort: Deleting group '%s'", channel->Name());
            Skins.Message(mtStatus, buf, messageTime);    
            }
	  dsyslog("AutoSort: Deleting group '%s'", channel->Name());
	  DeleteGroup(channel->Index());
	  channel = Channels.Get(saveIndex-1);
	  }
      }
    // Check Order
    //dsyslog("AutoSort: Checking order now");
    if (messageLevel >= 1) Skins.Message(mtStatus,"AutoSort: Checking Order of Group Delimiters", messageTime);
    group = AutoGroups.GetFirstByOrder();
    cAutoGroup *nextGroup = group;
    cChannel *channel = Channels.Get(group->GetChIndex());

    while ((nextGroup = AutoGroups.GetNextByOrder(group->OIdx()))){
      channel = Channels.Get(Channels.GetNextGroup(group->GetChIndex()));
      alarm(60);
      //dsyslog ("AutoSort: Have %s search %s found %s",group->Name(), nextGroup->Name(), channel->Name());
      if (strcmp(channel->Name(),nextGroup->Name()) == 0){
        group = nextGroup;
      } else {
        if (strcmp(Channels.Get(Channels.GetNextGroup(channel->Index()))->Name(),nextGroup->Name()) == 0){
	  //dsyslog("It's the overnext");
	  dsyslog("Moving Group '%s' to end",channel->Name());
	  if (messageLevel >= 2) {
            asprintf(&buf, "AutoSort: Moving '%s' to end", channel->Name());
            Skins.Message(mtStatus, buf, messageTime);    
            }
          //MoveGroupBehindGroup(nextGroup->GetChIndex(),group);	  
          MoveGroupToEnd(channel->Index());	            
	} else {
	  dsyslog("Moving Group '%s' behind '%s'",nextGroup->Name(),group->Name());
	  if (messageLevel >= 2) {
            asprintf(&buf, "AutoSort: Moving '%s' behind '%s'", nextGroup->Name(), group->Name());
            Skins.Message(mtStatus, buf, messageTime);    
            }
          MoveGroupBehindGroup(nextGroup->GetChIndex(),group);	  
          } // if overnext
	}
      } //while NextByOrder
     
  } //  if (!foundStart){
  if (messageLevel >= 1) Skins.Message(mtStatus,NULL,0);
}

void cNewAutoSort::MoveGroupBehindGroup (int ChIndex, cAutoGroup *group)
{
  cChannel *channel1 = Channels.Get(ChIndex);
  cChannel *channel2 = Channels.Get(Channels.GetNextGroup(ChIndex)-1);
  if (!channel2) channel2 = Channels.Last();
  int NextGroupIdx = Channels.GetNextGroup(group->GetChIndex());
  
  dsyslog("AutoSort: Moving Group '%s' behind '%s'",channel1->Name(),group->Name());
  if (channel1->Index() < AutoGroups.GetFirstByOrder()->GetChIndex() ){
    dsyslog("AutoSort: -ERROR- ::MoveGroupBehindGroup() tries working before Auto-Sort-Mark  -ABORTED-");
    return;
    }	    
  
  cAutoGroup *dummyGroup = new cAutoGroup;
  dummyGroup->SetName("Dummy");
  InsGroup(dummyGroup,NextGroupIdx);
  ChIndex++;
  
  cAutoGroup *dummyGroup2 = new cAutoGroup;
  dummyGroup2->SetName("Dummy2");
  InsGroup(dummyGroup2,NextGroupIdx);
  ChIndex++;

  if ( ChIndex == Channels.Last()->Index() ) {
    //dsyslog(" group %s is last and empty",channel1->Name());
    MoveChannelToGroup(ChIndex, dummyGroup);    
    }
  else if ( ChIndex == channel2->Index() ) {
    //dsyslog(" group %s is inbetween and empty",channel1->Name());
    MoveChannelToGroup(ChIndex, dummyGroup);    
    }
  else {    
    //dsyslog(" group %s is inbetween and filled",channel1->Name());
    int i =  channel2->Index();
    // ToDo: Make a heavy speed-up here
    //while (strcmp(channel1->Name(),Channels.Get(i)->Name()) !=0 ) {
    while (Channels.Get(i)->GroupSep() == false) {
      //alarm(60);
      MoveChannelToGroup(i, dummyGroup);
      }
    MoveChannelToGroup(i, dummyGroup);
  } // if
      
  channel1 = Channels.Get(dummyGroup->GetChIndex());    
  Channels.Del(channel1);
  Channels.ReNumber();
  channel1 = Channels.Get(dummyGroup2->GetChIndex());    
  Channels.Del(channel1);
  Channels.ReNumber();
  Channels.SetModified();
}

void cNewAutoSort::InsGroup (cAutoGroup *group, int ChIndex)
{
  if (strcmp(group->Name(),"Dummy") && strcmp(group->Name(),"Dummy2")){
    dsyslog("AutoSort: Inserting Group '%s'", group->Name());
    }
  cChannel *NewGroup = new cChannel;
  cChannel *Before = Channels.Get(ChIndex);
  char *name;
  asprintf(&name,":%s",group->Name());
  NewGroup->Parse(name);
  Channels.Ins(NewGroup,Before);
  Channels.ReNumber();
  Channels.SetModified();
}

void cNewAutoSort::MoveGroupToEnd (int ChIndex)
{
  cChannel *channel1 = Channels.Get(ChIndex);
  cChannel *channel2 = Channels.Get(Channels.GetNextGroup(ChIndex)-1);
  dsyslog("AutoSort: Moving Group %s to end",channel1->Name());
  
  cAutoGroup *group = new cAutoGroup;
  group->SetName("Dummy");
  AddGroup(group);

  cAutoGroup *group2 = new cAutoGroup;
  group2->SetName("Dummy2");
  AddGroup(group2);
  
  int i =  channel2->Index();
  while (i != ChIndex -1)
    {
    MoveChannelToGroup(i, group);
    i--;
    }
    
  channel1 = Channels.Get(group->GetChIndex());    
  Channels.Del(channel1);
  Channels.ReNumber();
  channel1 = Channels.Get(group2->GetChIndex());    
  Channels.Del(channel1);
  Channels.ReNumber();
  Channels.SetModified();    
}

void cNewAutoSort::AddGroup (cAutoGroup *group)
{
  dsyslog("AutoSort: Creating Group '%s'", group->Name());
  cChannel *NewGroup = new cChannel;
  char *name;
  
  if (group->FixedPos() > 0) {
    asprintf(&name, ":@%d %s",group->FixedPos(),group->Name());
  } else {
    asprintf(&name,":%s",group->Name());
    }
  
  NewGroup->Parse(name);
  Channels.Add(NewGroup);
  Channels.ReNumber();
  Channels.SetModified(); 
}


void cNewAutoSort::Dump(int from, int to)
{
  cChannel *channel = Channels.Get(from);
  while ((channel) && (channel->Index() <= to)) {
    if (channel->GroupSep()) {
      dsyslog ("-Group- %d: --%s--", channel->Index(), channel->Name());
    } else {
      dsyslog ("Channel %d:   %s", channel->Index(), channel->Name());
      }
    channel = Channels.Next(channel);
    }
}


void cNewAutoSort::DeleteGroup (int ChIndex)
{
  //Delete Group means delete the delimiter and move channels to Unsorted
  cChannel *channel1 = Channels.Get(ChIndex);
  cChannel *channel2;
  dsyslog("AutoSort: Deleting Group '%s'",channel1->Name());
  
  if (Channels.GetNextGroup(ChIndex) > 0){
    channel2 = Channels.Get(Channels.GetNextGroup(ChIndex)-1);
    cAutoGroup *group = AutoGroups.GetByName("Unsorted");
    for (int i = channel2->Index(); i > ChIndex; i--)
      {
      MoveChannelToGroup(i, group);
      }
    } 
  Channels.Del(channel1);
  Channels.ReNumber();
  Channels.SetModified();  
}


void cNewAutoSort::SetTimeStamp (int ChIndex)
{
  // TimeStamp is in seconds
  // 1 hour   = 3600
  // 1 day    = 86400
  // 1 week   = 604800
  // 1 month  = 2592000
  // 3 months = 7776000
    cChannel *channel = Channels.Get(ChIndex);
    if (!strstr(channel->ShortName(),"ASTS")){
      char *buf = NULL;
      int stampTime = time(NULL);
      asprintf(&buf,"ASTS-%d", stampTime);
      channel->SetName(channel->Name(), buf, channel->Provider());
    }
}

const char *cNewAutoSort::InGroup (int ChIndex)
{
  cChannel *channel = Channels.Get(Channels.GetPrevGroup(ChIndex));
  return channel->Name();
}

void cNewAutoSort::MoveChannelToGroup (int ChIndex, cAutoGroup *group)
{
  char *buf;
  cChannel *channel = Channels.Get(ChIndex);
  if (group->GetChIndex() == Channels.Last()->Index()){
    // Move channel behind last channel

    cAutoGroup *group2 = new cAutoGroup;
    group2->SetName("Dummy");
    AddGroup(group2);

    cChannel *ToChannel = Channels.Get(group->GetChIndex()+1);
    if (strcmp(group->Name(),"Dummy") != 0){
      dsyslog("AutoSort: move channel '%s' to group '%s'",channel->Name() ,group->Name());
      if (messageLevel >= 3) {
        asprintf(&buf, "AutoSort: Moving '%s' to '%s'", channel->Name(), group->Name());
        Skins.Message(mtStatus, buf, messageTime);    
        }
      }
    Channels.Move(channel, ToChannel);
    Channels.ReNumber();
    if (strcmp(InGroup(ToChannel->Index()), group->Name())!=0) {
      dsyslog ("ERROR: (MoveChannelToGroup1) Moved %s to %s instead of %s ", ToChannel->Name(), InGroup(ToChannel->Index()), group->Name());
      Dump(ToChannel->Index()-3,ToChannel->Index()+3);
      }

    cChannel *channel1 = Channels.Get(group2->GetChIndex());    
    //dsyslog("AutoSort: Deleting Group %s",channel1->Name());
    Channels.Del(channel1);
    Channels.ReNumber();
    Channels.SetModified();
  } else {
    // Move channel anywhere else
    cChannel *ToChannel;
    if (channel->Index() > group->GetChIndex()) {
      ToChannel = Channels.Get(group->GetChIndex()+1);
    } else {
      ToChannel = Channels.Get(group->GetChIndex());
      }
    if (strcmp(group->Name(),"Dummy") != 0){
      //dsyslog("move channel %d to %d (group %d)",channel->Index(), ToChannel->Index(), group->GetChIndex());
      dsyslog("AutoSort: move channel '%s' to group '%s'",channel->Name() ,group->Name());
      if (messageLevel >= 3) {
        asprintf(&buf, "AutoSort: Moving '%s' to '%s'", channel->Name(), group->Name());
        Skins.Message(mtStatus, buf, messageTime);    
        }
      }
    Channels.Move(channel, ToChannel);
    Channels.ReNumber();
    Channels.SetModified();
    //if (strcmp(InGroup(ToChannel->Index()), group->Name())!=0) {
      //dsyslog ("ERROR: (MoveChannelToGroup2) Moved %s to %s instead of %s ", ToChannel->Name(), InGroup(ToChannel->Index()), group->Name());
    if (strcmp(InGroup(channel->Index()), group->Name())!=0) { // XXX
      dsyslog ("ERROR: (MoveChannelToGroup2) Moved %s to %s instead of %s ", channel->Name(), InGroup(ToChannel->Index()), group->Name());
      Dump(ToChannel->Index()-3,ToChannel->Index()+3);
      }
  }
  // Sort here
  if (strcmp(group->Name(),"Dummy") != 0 ){
    Sort();
    }
}

// --------------------------------------------------------------

class cMenuAutoSortSetup : public cMenuSetupPage {
private:
  int newMaxMoves;
  int newMessageLevel;
  int newMessageTime;
  int newSortUnsorted;
  int newSortNew;
  int newSeperateNew;
  int newHideMenu;
  int newHousekeepingRun;
  int newVersionSorting;
  int newTimeStampSetDelta;
  int newWriteTimeStamp;
  int newUseFixedAuto;
  int newUseFixedUnsorted;
  int newUseFixedNew;
  int newFixedAuto;
  int newFixedUnsorted;
  int newFixedNew;
protected:
  virtual void Store(void);
public:
  cMenuAutoSortSetup(void);
  };

cMenuAutoSortSetup::cMenuAutoSortSetup(void)
{
  newTimeStampSetDelta = rint(timeStampSetDelta / 86400);
  newWriteTimeStamp    = writeTimeStamp;
  newMaxMoves          = maxMoves;
  newMessageLevel      = messageLevel;
  newMessageTime       = messageTime;
  newSeperateNew       = seperateNew;
  newSortNew           = sortNew;
  newSortUnsorted      = sortUnsorted;
  newHideMenu          = hideMenu;
  newHousekeepingRun   = housekeepingRun;
  newVersionSorting    = versionSorting;
  newUseFixedAuto      = useFixedAuto;
  newUseFixedUnsorted  = useFixedUnsorted;
  newUseFixedNew       = useFixedNew;
  newFixedAuto         = fixedAuto;
  newFixedUnsorted     = fixedUnsorted;
  newFixedNew          = fixedNew;
  Add(new cMenuEditBoolItem( "Hide MainMenu entry",             &newHideMenu));  
  Add(new cMenuEditBoolItem( "Run on housekeeping",             &newHousekeepingRun));  
  Add(new cMenuEditIntItem ( "Max channels to move",            &newMaxMoves));
  Add(new cMenuEditIntItem ( "Message level",                   &newMessageLevel, 0, 3));
  //Add(new cMenuEditIntItem ( "Message time (s)",                &newMessageTime));  
  Add(new cMenuEditBoolItem( "Use FixedPos for 'Auto Sort'",    &newUseFixedAuto));  
  Add(new cMenuEditIntItem ( "FixedPosition for 'Auto Sort'",   &newFixedAuto));
  Add(new cMenuEditBoolItem( "Sort 'Unsorted' channels",        &newSortUnsorted));  
  Add(new cMenuEditBoolItem( "Use FixedPos for 'Unsorted'",     &newUseFixedUnsorted));  
  Add(new cMenuEditIntItem ( "FixedPosition for 'Unsorted'",    &newFixedUnsorted));
  Add(new cMenuEditBoolItem( "Keep new channels seperated",     &newSeperateNew));  
  Add(new cMenuEditBoolItem( "Sort 'New Channels'",             &newSortNew));  
  Add(new cMenuEditBoolItem( "Use FixedPos for 'New Channels'", &newUseFixedNew));  
  Add(new cMenuEditIntItem ( "FixedPosition for 'New Channels'",&newFixedNew));
  Add(new cMenuEditBoolItem( "Use 'version' sort",              &newVersionSorting));  
  Add(new cMenuEditBoolItem( "Set TimeStamps",                  &newWriteTimeStamp));  
  Add(new cMenuEditIntItem ( "Days between setting TimeStamps", &newTimeStampSetDelta, 1));
}

void cMenuAutoSortSetup::Store(void)
{
  SetupStore("MaxMoves",        maxMoves          = newMaxMoves);
  SetupStore("MessageLevel",    messageLevel      = newMessageLevel);
  SetupStore("MessageTime",     messageTime       = newMessageTime);
  SetupStore("SeperateNew",     seperateNew       = newSeperateNew);
  SetupStore("SortNew",         sortNew           = newSortNew);
  SetupStore("SortUnsorted",    sortUnsorted      = newSortUnsorted);
  SetupStore("HousekeepingRun", housekeepingRun   = newHousekeepingRun);
  SetupStore("HideMenu",        hideMenu          = newHideMenu);
  SetupStore("VersionSort",     versionSorting    = newVersionSorting);
  SetupStore("SetTimeStamps",   writeTimeStamp    = newWriteTimeStamp);
  SetupStore("TimeStampDelta",  timeStampSetDelta = (newTimeStampSetDelta * 86400));
  SetupStore("UseFixedAuto",    useFixedAuto      = newUseFixedAuto);
  SetupStore("UseFixedUnsorted",useFixedUnsorted  = newUseFixedUnsorted);
  SetupStore("UseFixedNew",     useFixedNew       = newUseFixedNew);
  SetupStore("FixedAuto",       fixedAuto         = newFixedAuto);
  SetupStore("FixedUnsorted",   fixedUnsorted     = newFixedUnsorted);
  SetupStore("FixedNew",        fixedNew          = newFixedNew);
}

// --------------------------------------------------------------

class cPluginAutoSort : public cPlugin {
private:
  // Add any member variables or functions you may need here.
  cNewAutoSort *autoSort;
public:
  cPluginAutoSort(void);
  virtual ~cPluginAutoSort();
  virtual const char *Version(void) { return VERSION; }
  virtual const char *Description(void) { return DESCRIPTION; }
  virtual const char *CommandLineHelp(void);
  virtual bool ProcessArgs(int argc, char *argv[]);
  virtual bool Start(void);
  virtual void Housekeeping(void);
  virtual const char *MainMenuEntry(void) { if (!hideMenu) return MAINMENUENTRY; else return NULL; }
  virtual cOsdObject *MainMenuAction(void);
  virtual cMenuSetupPage *SetupMenu(void);
  virtual bool SetupParse(const char *Name, const char *Value);
  };

cPluginAutoSort::cPluginAutoSort(void)
{
  // Initialize any member variables here.
  // DON'T DO ANYTHING ELSE THAT MAY HAVE SIDE EFFECTS, REQUIRE GLOBAL
  // VDR OBJECTS TO EXIST OR PRODUCE ANY OUTPUT!
  autoSort = NULL;
}

cPluginAutoSort::~cPluginAutoSort()
{
  // Clean up after yourself!
  delete autoSort;
}

const char *cPluginAutoSort::CommandLineHelp(void)
{
  // Return a string that describes all known command line options.
  return NULL;
}

bool cPluginAutoSort::ProcessArgs(int argc, char *argv[])
{
  // Implement command line argument processing here if applicable.
  return true;
}

bool cPluginAutoSort::Start(void)
{
  // Start any background activities the plugin shall perform.
  autoSort = new cNewAutoSort;
  return true;
}

void cPluginAutoSort::Housekeeping(void)
{
  // Perform any cleanup or other regular tasks.
  if (housekeepingRun) autoSort->CheckChannels();
  if (saveSetup) {
    //dsyslog("save here!");
    SetupStore("LastTimeStamp", lastTimeStampSet);
    saveSetup = false;
    }
}

cOsdObject *cPluginAutoSort::MainMenuAction(void)
{
  // Perform the action when selected from the main VDR menu.
  delete autoSort;
  autoSort = new cNewAutoSort;
  //if (writeConfChange) SetupStore("LastConfChange",lastConfChange );
  dsyslog("AutoSort: Started complete run");  
  autoSort->CheckChannels(true);
  return NULL;
}
cMenuSetupPage *cPluginAutoSort::SetupMenu(void)
{
  // Return a setup menu in case the plugin supports one.
  return new cMenuAutoSortSetup;
}

bool cPluginAutoSort::SetupParse(const char *Name, const char *Value)
{
  // Parse your own setup parameters and store their values.
  if      (!strcasecmp(Name, "MaxMoves"))         maxMoves          = atoi(Value);
  else if (!strcasecmp(Name, "MessageLevel"))     messageLevel      = atoi(Value);
  else if (!strcasecmp(Name, "MessageTime"))      messageTime       = atoi(Value);
  else if (!strcasecmp(Name, "SeperateNew"))      seperateNew       = atoi(Value);
  else if (!strcasecmp(Name, "SortNew"))          sortNew           = atoi(Value);
  else if (!strcasecmp(Name, "SortUnsorted"))     sortUnsorted      = atoi(Value);
  else if (!strcasecmp(Name, "HousekeepingRun"))  housekeepingRun   = atoi(Value);
  else if (!strcasecmp(Name, "HideMenu"))         hideMenu          = atoi(Value);
  else if (!strcasecmp(Name, "VersionSort"))      versionSorting    = atoi(Value);  
  else if (!strcasecmp(Name, "LastTimeStamp"))    lastTimeStampSet  = atoi(Value);  
  else if (!strcasecmp(Name, "SetTimeStamps"))    writeTimeStamp    = atoi(Value);  
  else if (!strcasecmp(Name, "TimeStampDelta"))   timeStampSetDelta = atoi(Value);  
  else if (!strcasecmp(Name, "UseFixedAuto"))     useFixedAuto      = atoi(Value);
  else if (!strcasecmp(Name, "UseFixedUnsorted")) useFixedUnsorted  = atoi(Value);
  else if (!strcasecmp(Name, "UseFixedNew"))      useFixedNew       = atoi(Value);
  else if (!strcasecmp(Name, "FixedAuto"))        fixedAuto         = atoi(Value);
  else if (!strcasecmp(Name, "FixedUnsorted"))    fixedUnsorted     = atoi(Value);
  else if (!strcasecmp(Name, "FixedNew"))         fixedNew          = atoi(Value);
  else
     return false;
  return true;
}

VDRPLUGINCREATOR(cPluginAutoSort); // Don't touch this!
